#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libelf.h>
#include <gelf.h>

#define ERR(...) errx(EXIT_FAILURE, __VA_ARGS__)

// Input ELF
int in_fd;
Elf *in_elf;

// Output ELF
int out_fd;
Elf *out_elf;

void parse_ehdrs()
{
	GElf_Ehdr in_ehdr;
	Elf32_Ehdr *out_ehdr;
	int in_class;

	if (gelf_getehdr(in_elf, &in_ehdr) == NULL)
		ERR("getehdr() failed: %s\n", elf_errmsg(-1));

	if ((in_class = gelf_getclass(in_elf)) == ELFCLASSNONE)
		ERR("getclass() failed: %s\n", elf_errmsg(-1));

	if (in_class != ELFCLASS32)
		ERR("Sorry, 64 bit is not yet supported\n");

	if ((out_ehdr = elf32_newehdr(out_elf)) == NULL)
		ERR("elf32_newehdr() failed: %s\n", elf_errmsg(-1));

	strncpy(out_ehdr->e_ident, in_ehdr.e_ident, 16);
	out_ehdr->e_type		 = in_ehdr.e_type;
	out_ehdr->e_machine		 = in_ehdr.e_machine;
	out_ehdr->e_version		 = in_ehdr.e_version;
	out_ehdr->e_entry		 = in_ehdr.e_entry;
	out_ehdr->e_phoff		 = in_ehdr.e_phoff;
	out_ehdr->e_shoff		 = in_ehdr.e_shoff;
	out_ehdr->e_flags		 = in_ehdr.e_flags;
	out_ehdr->e_ehsize		 = in_ehdr.e_ehsize;
	out_ehdr->e_phentsize	 = in_ehdr.e_phentsize;
	out_ehdr->e_shentsize	 = in_ehdr.e_shentsize;

}

void parse_phdrs()
{
	GElf_Phdr in_phdr;
	Elf32_Phdr *out_phdrs;
	size_t phdr_count;
	int i;


	if (elf_getphdrnum(in_elf, &phdr_count) != 0)
		ERR("elf_getphdrnum() failed: %s\n", elf_errmsg(-1));

	if (phdr_count == 0)
		return;
	
	if ((out_phdrs = elf32_newphdr(out_elf, phdr_count)) == NULL)
		ERR("elf32_newphdr() failed: %s\n", elf_errmsg(-1));

	for (i = 0; i < phdr_count; ++i) 
	{
		if (gelf_getphdr(in_elf, i, &in_phdr) != &in_phdr)
			ERR("getphdr() failed: %s\n", elf_errmsg(-1));

		out_phdrs[i].p_type		 = in_phdr.p_type;
		out_phdrs[i].p_offset	 = in_phdr.p_offset;
		out_phdrs[i].p_vaddr	 = in_phdr.p_vaddr;
		out_phdrs[i].p_paddr	 = in_phdr.p_paddr;
		out_phdrs[i].p_filesz	 = in_phdr.p_filesz;
		out_phdrs[i].p_memsz	 = in_phdr.p_memsz;
		out_phdrs[i].p_flags	 = in_phdr.p_flags;
		out_phdrs[i].p_align	 = in_phdr.p_align;

	}

}

void parse_scns(const char* scn_name, const char* patch_path)
{
	Elf_Scn *in_scn;
	Elf_Scn *out_scn;

	Elf_Data *in_data;
	Elf_Data *out_data;

	GElf_Shdr in_shdr;
	Elf32_Shdr *out_shdr;
	
	char* in_name;
	size_t shstrndx;
	int offset_correction;


	if (elf_getshdrstrndx(in_elf, &shstrndx) != 0)
		ERR("elf_getshdrstrndx() failed: %s\n", elf_errmsg(-1));

	in_scn = NULL;
	offset_correction = 0;

	while ((in_scn = elf_nextscn(in_elf, in_scn)) != NULL)
	{
		// Input
		in_data = NULL;

		if (gelf_getshdr(in_scn, &in_shdr) != &in_shdr)
			ERR("getshdr() failed: %s\n", elf_errmsg(-1));
		
		if ((in_data = elf_getdata(in_scn, in_data)) == NULL)
			ERR("elf_getdata() failed: %s\n", elf_errmsg(-1));

		// Output
		if ((out_scn = elf_newscn(out_elf)) == NULL)
			ERR("elf_newscn() failed: %s\n", elf_errmsg(-1));

		if ((out_data = elf_newdata(out_scn)) == NULL)
			ERR("elf_newdata() failed: %s\n", elf_errmsg(-1));

		out_data->d_buf 		= in_data->d_buf;
		out_data->d_type 		= in_data->d_type;
		out_data->d_size 		= in_data->d_size;
		out_data->d_off 		= in_data->d_off;
		out_data->d_align 		= in_data->d_align;
		out_data->d_version 	= in_data->d_version;

		if ((out_shdr = elf32_getshdr(out_scn)) == NULL)
			ERR("elf32_getshdr() failed: %s\n", elf_errmsg(-1));

		out_shdr->sh_name		 = in_shdr.sh_name;
		out_shdr->sh_type		 = in_shdr.sh_type;
		out_shdr->sh_flags		 = in_shdr.sh_flags;
		out_shdr->sh_addr		 = in_shdr.sh_addr;
		out_shdr->sh_offset		 = in_shdr.sh_offset + offset_correction;
		out_shdr->sh_size		 = in_shdr.sh_size;
		out_shdr->sh_link		 = in_shdr.sh_link;
		out_shdr->sh_info		 = in_shdr.sh_info;
		out_shdr->sh_addralign	 = 
			(in_shdr.sh_addralign > 0) ? in_shdr.sh_addralign : 1;
		out_shdr->sh_entsize	 = in_shdr.sh_entsize;

		// Check section name
		if ((in_name = elf_strptr(in_elf, shstrndx, in_shdr.sh_name)) == NULL)
			ERR("elf_strptr() failed: %s\n", elf_errmsg(-1));
		
		if (!strncmp(in_name, scn_name, strlen(in_name)))
		{
			FILE* p_fd = fopen(patch_path, "r");
			unsigned char* p_buffer;
			long int p_size;

			// Get the size of the patch data
			fseek(p_fd, 0, SEEK_END);
			p_size = ftell(p_fd);
			rewind(p_fd);

			// Read the patch data
			p_buffer = calloc(p_size, sizeof(unsigned char));
			fread(p_buffer, sizeof(unsigned char), p_size, p_fd);

			out_data->d_buf = p_buffer;
			out_data->d_size = p_size;
			out_shdr->sh_size = p_size;

			offset_correction = (p_size - in_data->d_size);

			fclose(p_fd);
		}

	}

}

int main(int argc, char **argv)
{
	const char *in_file_path;
	const char *section_name;
	const char *patch_path;
	const char *out_file_path;

	if (argc != 5) 
		ERR("Usage: %s <input> <section> <patch> <output>\n", argv[0]);

	in_file_path = argv[1];
	section_name = argv[2];
	patch_path = argv[3];
	out_file_path = argv[4];

	// libelf version check, init
	if (elf_version(EV_CURRENT) == EV_NONE)
		ERR("ELF library initialization failed: %s\n", elf_errmsg(-1));

	// File descriptors
	if ((in_fd = open(in_file_path, O_RDONLY, 0)) < 0)
		ERR("Open \"%s\" failed\n", in_file_path);

	if ((out_fd = open(out_file_path, O_WRONLY | O_CREAT, 0777)) < 0)
		ERR("Open \"%s\" failed\n", out_file_path);

	// ELF object handles
	if ((in_elf = elf_begin(in_fd, ELF_C_READ, NULL)) == NULL)
		ERR("elf_begin() failed: %s\n", elf_errmsg(-1));

	if ((out_elf = elf_begin(out_fd, ELF_C_WRITE, NULL)) == NULL)
		ERR("elf_begin() failed: %s\n", elf_errmsg(-1));

	// ELF object type check
	if (elf_kind(in_elf) != ELF_K_ELF)
		ERR("\"%s\" is not an ELF object\n", in_file_path);

	// TODO: 
	//		Copy everything from in_ef to out_elf
	//		With the exception of the section with the 
	//		name of the given parameter


	// Set ELF_F_LAYOUT flag
	elf_flagelf(out_elf, ELF_C_SET, ELF_F_LAYOUT);

	// ELF header
	parse_ehdrs();

	// Program header
	parse_phdrs();

	// Section header
	parse_scns(section_name, patch_path);

	if (elf_update(out_elf, ELF_C_WRITE) < 0)
		ERR("elf_update() failed: %s\n", elf_errmsg(-1));

	elf_end(out_elf);
	elf_end(in_elf);

	close(out_fd);
	close(in_fd);

	exit(EXIT_SUCCESS);
}